/*
 * File:   main.c
 * Author: zacharymurtishi
 *
 * Created on October 3, 2024, 5:31 PM
 */


// PIC18F2420 Configuration Bit Settings

// 'C' source line config statements

// CONFIG1H
#pragma config OSC = INTIO67    // Oscillator Selection bits (Internal oscillator block, port function on RA6 and RA7)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enable bit (Fail-Safe Clock Monitor disabled)
#pragma config IESO = OFF       // Internal/External Oscillator Switchover bit (Oscillator Switchover mode disabled)

// CONFIG2L
#pragma config PWRT = ON        // Power-up Timer Enable bit (PWRT enabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable bits (Brown-out Reset disabled in hardware and software)
#pragma config BORV = 3         // Brown Out Reset Voltage bits (Minimum setting)

// CONFIG2H
#pragma config WDT = OFF        // Watchdog Timer Enable bit (WDT disabled (control is placed on the SWDTEN bit))
#pragma config WDTPS = 32768    // Watchdog Timer Postscale Select bits (1:32768)

// CONFIG3H
#pragma config CCP2MX = PORTC   // CCP2 MUX bit (CCP2 input/output is multiplexed with RC1)
#pragma config PBADEN = OFF     // PORTB A/D Enable bit (PORTB<4:0> pins are configured as digital I/O on Reset)
#pragma config LPT1OSC = OFF    // Low-Power Timer1 Oscillator Enable bit (Timer1 configured for higher power operation)
#pragma config MCLRE = ON       // MCLR Pin Enable bit (MCLR pin enabled; RE3 input pin disabled)

// CONFIG4L
#pragma config STVREN = OFF     // Stack Full/Underflow Reset Enable bit (Stack full/underflow will not cause Reset)
#pragma config LVP = OFF        // Single-Supply ICSP Enable bit (Single-Supply ICSP disabled)
#pragma config XINST = OFF      // Extended Instruction Set Enable bit (Instruction set extension and Indexed Addressing mode disabled (Legacy mode))

// CONFIG5L
#pragma config CP0 = OFF        // Code Protection bit (Block 0 (000800-001FFFh) not code-protected)
#pragma config CP1 = OFF        // Code Protection bit (Block 1 (002000-003FFFh) not code-protected)

// CONFIG5H
#pragma config CPB = OFF        // Boot Block Code Protection bit (Boot block (000000-0007FFh) not code-protected)
#pragma config CPD = OFF        // Data EEPROM Code Protection bit (Data EEPROM not code-protected)

// CONFIG6L
#pragma config WRT0 = OFF       // Write Protection bit (Block 0 (000800-001FFFh) not write-protected)
#pragma config WRT1 = OFF       // Write Protection bit (Block 1 (002000-003FFFh) not write-protected)

// CONFIG6H
#pragma config WRTC = OFF       // Configuration Register Write Protection bit (Configuration registers (300000-3000FFh) not write-protected)
#pragma config WRTB = OFF       // Boot Block Write Protection bit (Boot block (000000-0007FFh) not write-protected)
#pragma config WRTD = OFF       // Data EEPROM Write Protection bit (Data EEPROM not write-protected)

// CONFIG7L
#pragma config EBTR0 = OFF      // Table Read Protection bit (Block 0 (000800-001FFFh) not protected from table reads executed in other blocks)
#pragma config EBTR1 = OFF      // Table Read Protection bit (Block 1 (002000-003FFFh) not protected from table reads executed in other blocks)

// CONFIG7H
#pragma config EBTRB = OFF      // Boot Block Table Read Protection bit (Boot block (000000-0007FFh) not protected from table reads executed in other blocks)

// #pragma config statements should precede project file includes.
// Use project enums instead of #define for ON and OFF.

#include <xc.h>

#define LED_IC_ADDR_WR 0b11000000
#define LED_IC_ADDR_RD 0b11000001

#define LEDOUT_START 0x94
#define PWM_LED_ARR_START 0x82

#define _XTAL_FREQ 4000000

#define BUTTON_SCAN0 PORTAbits.RA0
#define BUTTON_SCAN1 PORTAbits.RA1
#define BUTTON_SCAN2 PORTAbits.RA2
#define BUTTON_SCAN3 PORTAbits.RA3

#define BUTTON_READ0 PORTBbits.RB0
#define BUTTON_READ1 PORTBbits.RB1
#define BUTTON_READ2 PORTBbits.RB2
#define BUTTON_READ3 PORTBbits.RB3
#define PLAYER1 0
#define PLAYER2 1
#define MAX_LEDS 16

#define MANCALA0 gameTable[6]
#define MANCALA1 gameTable[13]

uint8_t ledBrightness[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
uint8_t pots_player1[6] = {0,0,0,0,0,0};
uint8_t pots_player2[6] = {0,0,0,0,0,0};
uint8_t complimentaryPots1[6] = {12,11,10,9,8,7};
uint8_t complimentaryPots2[6] = {5,4,3,2,1,0};
uint8_t gameTable[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
uint8_t player = PLAYER1;
uint8_t scan = 0;
uint8_t pressed = 0;
uint16_t ticks = 0;
uint8_t brightness = 0x0E;

void wait(uint16_t ms){
    ticks=0;
    while(ticks<ms);
    return;
}

uint8_t debounce(uint8_t button, uint8_t buttonNumber){
    wait(5);
    if((PORTB>>button)&0b1){
        while(((PORTB>>button)&0b1))
        {
            ticks=0;
        }
        return buttonNumber;
    }    
    else{
        return 0;
    }
}

void __interrupt() isr(void){
    if(INTCONbits.TMR0IF){
        TMR0IF = 0;
        TMR0 = 65532;
        ticks++;
        if(((ticks % 20)) == 0){
            scan = (scan+1)&0b00000011;
        }
        if(scan == 0){
            PORTA = 0b00000001;
        }
        else if(scan == 1){
            PORTA = 0b00000010;
        }
        else if(scan == 2){
            PORTA = 0b00000100;
        }
        else if(scan == 3){
            PORTA = 0b00001000;
        }         
        return;
    }
}

void Timer0_Init(void){
    T0CONbits.TMR0ON = 1;
    T0CONbits.T0CS = 0;
    T0CONbits.T08BIT = 0;
    T0CONbits.PSA = 0;
    T0CONbits.T0PS = 0b111;
    TMR0 = 65532;
}

void interruptsEnable(void){
    INTCONbits.PEIE = 1;
    INTCONbits.GIE = 1;
    INTCONbits.T0IE = 1;
}

void interruptsDisable(void){
    INTCONbits.PEIE = 0;
    INTCONbits.GIE = 0;
    INTCONbits.T0IE = 0;
}

void I2C_Idle(void){
    while((SSPSTAT & 0x04) || (SSPCON2 & 0x1F));
}

void I2C_Start(void){
    SSPCON2bits.SEN = 1;
    while(SSPCON2bits.SEN);
}

void I2C_Repeated_Start(void){
    SSPCON2bits.RSEN = 1;
    while(SSPCON2bits.RSEN);
}

void I2C_Init(void){
    //TRISA = TRISA | 0b11111111;
    SSPCON1bits.SSPEN = 1;
    SSPADD = 9;
    SSPCON1bits.SSPM = 0b1000;
}

void I2C_Send(uint8_t byte){
    SSPCON2bits.ACKSTAT = 0;
    SSPCON2bits.RCEN = 0;
    I2C_Idle();
    SSPBUF = byte;
    while(SSPCON1bits.WCOL){
       SSPCON1bits.WCOL = 0;
       SSPBUF = byte;
    }
    while(SSPSTATbits.BF);
    while(SSPCON2bits.ACKSTAT);
    SSPCON2bits.ACKSTAT = 0;
}


uint8_t I2C_Receive_ACK(){
    uint8_t received_byte;
    SSPCON2bits.RCEN = 1;
    //received_byte = SSPBUF;
    while(SSPSTATbits.BF == 0);
    received_byte = SSPBUF;
    SSPCON2bits.ACKDT = 0;
    SSPCON2bits.ACKEN = 1;
    while(SSPCON2bits.ACKEN);
    return received_byte;
}

uint8_t I2C_Receive_NACK(){
    uint8_t received_byte;
    //received_byte = SSPBUF;
    SSPCON2bits.RCEN = 1;
    while(SSPSTATbits.BF == 0);
    received_byte = SSPBUF;
    SSPCON2bits.ACKDT = 1;
    SSPCON2bits.ACKEN = 1;
    while(SSPCON2bits.ACKEN);
    return received_byte;
}

void I2C_Stop(void){
    SSPCON2bits.PEN = 1;
    while(SSPCON2bits.PEN);
}

void initLEDs(void){
    I2C_Init();
    for(uint8_t i=0; i<MAX_LEDS; i++){
        ledBrightness[i] = 0;
    }
    I2C_Start();
    I2C_Send(LED_IC_ADDR_WR);
    I2C_Send(0x00);
    I2C_Send(0x00);
    I2C_Stop();    
    
    I2C_Start();
    I2C_Send(LED_IC_ADDR_WR);
    I2C_Send(LEDOUT_START);
    for(uint8_t i=0; i<4; i++){
        I2C_Send(0xAA);
    }
    I2C_Stop();
    
    I2C_Start();
    I2C_Send(LED_IC_ADDR_WR);
    I2C_Send(0x1C);
    I2C_Send(0b01001111);
    I2C_Stop();    
}

void setLED(uint8_t LED_ADDR){
    uint8_t address = LED_ADDR + 0x02;
    I2C_Start();
    I2C_Send(LED_IC_ADDR_WR);
    I2C_Send(address);
    I2C_Send(brightness);
    I2C_Stop();
}

void setLEDs(void){
    brightness = 0x0E;
    I2C_Start();
    I2C_Send(LED_IC_ADDR_WR);
    I2C_Send(PWM_LED_ARR_START);
    for(uint8_t i=0; i<MAX_LEDS; i++){
        I2C_Send(brightness);
    }
    I2C_Stop();
}


void clearLEDs(void){
    I2C_Start();
    I2C_Send(LED_IC_ADDR_WR);
    I2C_Send(PWM_LED_ARR_START);
    for(uint8_t i=0; i<MAX_LEDS; i++){
        I2C_Send(0);
    }
    I2C_Stop();
}

void gameOver(void){
    if(MANCALA1 > MANCALA0){
        while(1){
            clearLEDs();
            wait(1000);
            setLED(14);
            wait(1000);
        }
    }
    else if(MANCALA0 > MANCALA1){
        while(1){
            clearLEDs();
            wait(1000);
            setLED(15);
            wait(1000);
        }
    }
    else if(MANCALA0 == MANCALA1){
        while(1){
            clearLEDs();
            wait(1000);
            setLED(14);
            setLED(15);
            wait(1000);
        }
    }
}

void updateLEDs(void){
    
    for(uint8_t i=0; i<14; i++)
    {
        if(gameTable[i]){
            ledBrightness[i] = brightness;
        }
        else{
            ledBrightness[i] = 0;
        }
    }   
    if(!player){
        ledBrightness[14] = brightness;
        ledBrightness[15] = 0;
    }
    else{
        ledBrightness[14] = 0;
        ledBrightness[15] = brightness;            
        }
    I2C_Start();
    I2C_Send(LED_IC_ADDR_WR);
    I2C_Send(PWM_LED_ARR_START);
    for(uint8_t i=0; i<MAX_LEDS; i++){
        I2C_Send(ledBrightness[i]);
    }
    I2C_Stop();
}

uint8_t scanForButtons(void){
    uint8_t scanned = 0;
    while(1){
        if(BUTTON_READ0 && (scan==0)){
            scanned = debounce(0, 4);
        }
        else if(BUTTON_READ1 && (scan==0)){
            scanned = debounce(1, 8);
        }
        else if(BUTTON_READ2 && (scan==0)){
            scanned = debounce(2, 12);
        }
        else if(BUTTON_READ3 && (scan==0)){
            scanned = debounce(3, 16);
        }
        if(BUTTON_READ0 && (scan==1)){
            scanned = debounce(0, 3);
        }
        else if(BUTTON_READ1 && (scan==1)){
            scanned = debounce(1, 7);
        }
        else if(BUTTON_READ2 && (scan==1)){
            scanned = debounce(2, 11);
        }
        else if(BUTTON_READ3 && (scan==1)){
            scanned = debounce(3, 15);
        }
        if(BUTTON_READ0 && (scan==2)){
            scanned = debounce(0, 2);
        }
        else if(BUTTON_READ1 && (scan==2)){
            scanned = debounce(1, 6);
        }
        else if(BUTTON_READ2 && (scan==2)){
            scanned = debounce(2, 10);
        }
        else if(BUTTON_READ3 && (scan==2)){
            scanned = debounce(3, 14);
        }  
        if(BUTTON_READ0 && (scan==3)){
            scanned = debounce(0, 1);
        }
        else if(BUTTON_READ1 && (scan==3)){
            scanned = debounce(1, 5);
        }
        else if(BUTTON_READ2 && (scan==3)){
            scanned = debounce(2, 9);
        }
        else if(BUTTON_READ3 && (scan==3)){
            scanned = debounce(3, 13);
        }
        if(scanned != 0){
            break;
        }
    }
    return scanned;
}

void initMatrix(void){
    TRISA = 0b11110000;
    TRISB = 0xFF; 
}

void brightnessAdjust(void){
    for(uint8_t i=0; i<MAX_LEDS; i++){
        if(ledBrightness[i])
        {
            ledBrightness[i] = brightness;
        }
    }
    updateLEDs();
}

void initGameState(void){
    brightness = 0x0E;
    for(uint8_t i=0; i<14; i++)
    {
        gameTable[i] = 4;
    }  
    MANCALA0 = 0;
    MANCALA1 = 0;
    for(uint8_t i=0; i<14; i++)
    {
        if(gameTable[i]){
            ledBrightness[i] = brightness;
        }
        else{
            ledBrightness[i] = 0;
        }
    }
    player = PLAYER1;
    ledBrightness[14] = brightness;
    ledBrightness[15] = 0;
    updateLEDs();
    
}

uint8_t isEmpty(uint8_t *arr, uint8_t len){
    for(uint8_t i=0; i<len; i++){
        if(arr[i]){
            return 0;
        }
    }
    return 1;
}

uint8_t sum(uint8_t *arr, uint8_t len){
    uint8_t sum = 0;
    for(uint8_t i=0; i<len; i++){
        sum = sum + arr[i];
    }
    return sum;
}

void makeMove(uint8_t start){
    uint8_t skip = 0;
    uint8_t addr = 0;
    uint8_t last;
    for(uint8_t i = (start+238); i>((start-gameTable[start])+238); i--){
        if(player){
            if(i%14 == 13){
                skip++;
            }
        }
        else{   
            if(i%14 == 6){
                skip++;
            }            
        }
        addr = (i-skip)%14;
        gameTable[addr]++;
        clearLEDs();
        wait(250);
        setLED(addr);
        wait(250);
    } 
    if(!player){
        if(!(addr == 13)){
            player = player ^ 1;
            if((gameTable[addr] == 1) && ((addr >= 0) && (addr <= 5))){
                MANCALA1 = MANCALA1 + gameTable[complimentaryPots1[addr]];
                MANCALA1++;
                gameTable[addr] = 0;
                if(gameTable[complimentaryPots1[addr]]){
                    for(uint8_t i=0; i<3; i++){
                        clearLEDs();
                        wait(100);
                        setLED(addr);
                        setLED(complimentaryPots1[addr]);
                        wait(100);
                    }                 
                }
                gameTable[complimentaryPots1[addr]] = 0;
            }
        }
        else{
             for(uint8_t i=0; i<3; i++){
                clearLEDs();
                wait(100);
                setLED(13);
                setLED(14);
                wait(100);
            }           
        }
    }
    else{   
        if(!(addr == 6)){
            player = player ^ 1;
            if((gameTable[addr] == 1) && ((addr >= 7) && (addr <= 12))){
                MANCALA0 = MANCALA0 + gameTable[complimentaryPots2[(addr-1)%6]];
                MANCALA0++;
                gameTable[addr] = 0;
                if(gameTable[complimentaryPots2[(addr-1)%6]]){
                    for(uint8_t i=0; i<3; i++){
                        clearLEDs();
                        wait(100);
                        setLED(addr);
                        setLED(complimentaryPots2[(addr-1)%6]);
                        wait(100);
                    }                
                    gameTable[complimentaryPots2[(addr-1)%6]] = 0;
                }
            }            
        }
        else{
            for(uint8_t i=0; i<3; i++){
                clearLEDs();
                wait(100);
                setLED(6);
                setLED(15);
                wait(100);
            }
        }
    }   
    gameTable[start] = 0;    
    for(uint8_t j=0; j<6; j++){
        pots_player1[j] = gameTable[j];
        pots_player2[j] = gameTable[j+7];        
    }
    if(isEmpty(pots_player1, sizeof(pots_player1))){
        for(uint8_t i=0; i<6; i++){
            if(gameTable[i+7]){
                setLED(i+7);
            }
        }        
        wait(3000);
        MANCALA1 = MANCALA1 + sum(pots_player2, sizeof(pots_player2));
        gameOver();
    }
    else if(isEmpty(pots_player2, sizeof(pots_player2))){
        for(uint8_t i=0; i<6; i++){
            if(gameTable[i]){
                setLED(i);
            }
        }
        wait(3000);
        MANCALA0 = MANCALA0 + sum(pots_player1, sizeof(pots_player1));
        gameOver();
    }    
    wait(1000);
    updateLEDs();
}

void howMany(uint8_t pot){
    clearLEDs();
    for(uint8_t i=0; i<gameTable[pot]; i++){
        wait(250);
        setLED(pot);
        wait(250);       
        clearLEDs();
    }
    wait(500);        
    updateLEDs();
}

void lampTest(void){
    for(uint8_t i=0; i<MAX_LEDS; i++){
        setLED(i);
        wait(100);
    }
    wait(250);
    clearLEDs();
    wait(250);
}

void main(void) {
    OSCCONbits.IRCF = 0b110;
    uint8_t activeButton = 0;
    Timer0_Init();    
    interruptsEnable();    
    initLEDs();
    initMatrix();
    lampTest();
    initGameState();
    updateLEDs();
    uint8_t selected = 0xFF;
    while(1){
        activeButton = scanForButtons() - 1;
        if(activeButton == 14){
            
            if(brightness<0xFF){
                brightness = brightness + 1;
            }
            brightnessAdjust();
        }
        else if(activeButton == 15){
            if(brightness>0x04){
                brightness = brightness - 1;              
            }          
            brightnessAdjust();
        }
        else if(activeButton == 6){
            if(gameTable[activeButton]){
                howMany(activeButton);
            }
            selected = 255;
        }
        else if(activeButton == 13){
            if(gameTable[activeButton]){
                howMany(activeButton);
            }   
            selected = 255;
        }    
        else if(activeButton == selected){
            if(((activeButton >= 0) && (activeButton <= 6)) && (!player)){
                makeMove(activeButton);
                selected = 255;
            }
            else if(((activeButton >= 7) && (activeButton <= 12)) && player){
                makeMove(activeButton);
                selected = 255;
            }       
            else{
                if(gameTable[activeButton]){
                    howMany(activeButton);
                }                
            }
        }           
        else{
            selected = activeButton;
            if(gameTable[activeButton]){
                howMany(activeButton);
            }
        }
    }
}